@@ -26,9 +26,10 @@
 // ********************* Private part **************************
 
 #define MAX_DIRTS           16
+#define EVDI_INVALID_DEVICE_INDEX -1
 
 #define EVDI_MODULE_COMPATIBILITY_VERSION_MAJOR 1
-#define EVDI_MODULE_COMPATIBILITY_VERSION_MINOR 5
+#define EVDI_MODULE_COMPATIBILITY_VERSION_MINOR 9
 #define EVDI_MODULE_COMPATIBILITY_VERSION_PATCH 0
 
 #define evdi_log(...) do {						\
@@ -58,6 +59,9 @@ struct evdi_device_context {
 	int device_index;
 };
 
+#define EVDI_USAGE_LEN 64
+static evdi_handle card_usage[EVDI_USAGE_LEN];
+
 static int drm_ioctl(int fd, unsigned long request, void *arg)
 {
 	int ret;
@@ -426,6 +430,146 @@ static int open_device(int device)
 	return fd;
 }
 
+
+static int write_add_device(const char *buffer, size_t buffer_length)
+{
+	FILE *add_devices = fopen("/sys/devices/evdi/add", "w");
+	int written = 0;
+
+	if (add_devices != NULL) {
+		written = fwrite(buffer,
+				 1,
+				 buffer_length,
+				 add_devices);
+		fclose(add_devices);
+	}
+
+	return written;
+}
+
+static int get_drm_device_index(const char *evdi_sysfs_drm_dir)
+{
+	struct dirent *fd_entry;
+	DIR *fd_dir;
+	int dev_index = EVDI_INVALID_DEVICE_INDEX;
+
+	fd_dir = opendir(evdi_sysfs_drm_dir);
+	if (fd_dir == NULL) {
+		evdi_log("Failed to open dir %s", evdi_sysfs_drm_dir);
+		return dev_index;
+	}
+
+	while ((fd_entry = readdir(fd_dir)) != NULL) {
+		if (strncmp(fd_entry->d_name, "card", 4) == 0)
+			dev_index = strtol(&fd_entry->d_name[4], NULL, 10);
+	}
+	closedir(fd_dir);
+
+	return dev_index;
+}
+
+static bool is_correct_parent_device(const char *dirname, const char *parent_device)
+{
+	char link_path[PATH_MAX];
+
+	snprintf(link_path, PATH_MAX - 7, "%s/device", dirname);
+
+	if (parent_device == NULL)
+		return access(link_path, F_OK) != 0;
+
+	char link_resolution[PATH_MAX];
+	const ssize_t link_resolution_len = readlink(link_path, link_resolution, PATH_MAX);
+
+	if (link_resolution_len == -1 || link_resolution_len == PATH_MAX)
+		return false;
+
+	link_resolution[link_resolution_len] = '\0';
+	char *parent_device_token = strrchr(link_resolution, '/');
+
+	if (strlen(parent_device) < 2)
+		return false;
+
+	parent_device_token++;
+	size_t len = strlen(parent_device_token);
+
+	bool is_same_device = strlen(parent_device) == len && strncmp(parent_device_token, parent_device, len) == 0;
+
+	return is_same_device;
+}
+
+static int find_unused_card_for(const char *parent_device)
+{
+	char evdi_platform_root[] = "/sys/bus/platform/devices";
+	struct dirent *fd_entry;
+	DIR *fd_dir;
+	int device_index = EVDI_INVALID_DEVICE_INDEX;
+
+	fd_dir = opendir(evdi_platform_root);
+	if (fd_dir == NULL) {
+		evdi_log("Failed to open dir %s", evdi_platform_root);
+		return device_index;
+	}
+
+	while ((fd_entry = readdir(fd_dir)) != NULL) {
+		if (strncmp(fd_entry->d_name, "evdi", 4) != 0)
+			continue;
+
+		char evdi_path[PATH_MAX];
+
+		snprintf(evdi_path, PATH_MAX, "%s/%s", evdi_platform_root, fd_entry->d_name);
+
+		if (!is_correct_parent_device(evdi_path, parent_device))
+			continue;
+
+		char evdi_drm_path[PATH_MAX];
+
+		snprintf(evdi_drm_path, PATH_MAX - strlen(evdi_path), "%s/drm", evdi_path);
+		int dev_index = get_drm_device_index(evdi_drm_path);
+
+		assert(dev_index < EVDI_USAGE_LEN && dev_index >= 0);
+
+		if (card_usage[dev_index] == EVDI_INVALID_HANDLE) {
+			device_index = dev_index;
+			break;
+		}
+	}
+	closedir(fd_dir);
+
+	return device_index;
+}
+
+static int get_generic_device(void)
+{
+	int device_index = EVDI_INVALID_DEVICE_INDEX;
+
+	device_index = find_unused_card_for(NULL);
+	if (device_index == EVDI_INVALID_DEVICE_INDEX) {
+		evdi_log("Creating card in /sys/devices/platform");
+		write_add_device("1", 1);
+		device_index = find_unused_card_for(NULL);
+	}
+
+	return device_index;
+}
+
+static int get_device_attached_to_usb(const char *sysfs_parent_device)
+{
+	int device_index = EVDI_INVALID_DEVICE_INDEX;
+	const char *parent_device = &sysfs_parent_device[4];
+
+	device_index = find_unused_card_for(parent_device);
+	if (device_index == EVDI_INVALID_DEVICE_INDEX) {
+		evdi_log("Creating card for usb device %s in /sys/bus/platform/devices", parent_device);
+		const size_t len = strlen(sysfs_parent_device);
+
+		write_add_device(sysfs_parent_device, len);
+		device_index = find_unused_card_for(parent_device);
+	}
+
+	return device_index;
+}
+
+
 // ********************* Public part **************************
 
 evdi_handle evdi_open(int device)
@@ -440,6 +584,8 @@ evdi_handle evdi_open(int device)
 			if (h) {
 				h->fd = fd;
 				h->device_index = device;
+				card_usage[device] = h;
+				evdi_log("Using /dev/dri/card%d", device);
 			}
 		}
 		if (h == EVDI_INVALID_HANDLE)
@@ -458,7 +604,7 @@ static enum evdi_device_status evdi_device_to_platform(int device, char *path)
 	if (!device_exists(device))
 		return NOT_PRESENT;
 
-	fd_dir = opendir("/sys/devices/platform");
+	fd_dir = opendir("/sys/bus/platform/devices");
 	if (fd_dir == NULL) {
 		evdi_log("Failed to list platform devices");
 		return NOT_PRESENT;
@@ -469,7 +615,7 @@ static enum evdi_device_status evdi_device_to_platform(int device, char *path)
 			continue;
 
 		snprintf(path, PATH_MAX,
-			"/sys/devices/platform/%s", fd_entry->d_name);
+			"/sys/bus/platform/devices/%s", fd_entry->d_name);
 		snprintf(card_path, PATH_MAX, "%s/drm/card%d", path, device);
 		if (path_exists(card_path)) {
 			status = AVAILABLE;
@@ -490,42 +636,55 @@ enum evdi_device_status evdi_check_device(int device)
 
 int evdi_add_device(void)
 {
-	FILE *add_devices = fopen("/sys/devices/evdi/add", "w");
-	int written = 0;
+	return write_add_device("1", 1);
+}
 
-	if (add_devices != NULL) {
-		static const char devices_to_add[] = "1";
-		const size_t elem_bytes = 1;
+evdi_handle evdi_open_attached_to(const char *sysfs_parent_device)
+{
+	int device_index = EVDI_INVALID_DEVICE_INDEX;
 
-		written = fwrite(devices_to_add,
-				 elem_bytes,
-				 sizeof(devices_to_add),
-				 add_devices);
-		fclose(add_devices);
+	if (sysfs_parent_device == NULL)
+		device_index = get_generic_device();
+
+	if (sysfs_parent_device != NULL && strncmp(sysfs_parent_device, "usb:", 4) == 0 && strlen(sysfs_parent_device) > 4)
+		device_index = get_device_attached_to_usb(sysfs_parent_device);
+
+	if (device_index >= 0 && device_index < EVDI_USAGE_LEN)  {
+		evdi_handle handle = evdi_open(device_index);
+		return handle;
 	}
 
-	return written;
+	return EVDI_INVALID_HANDLE;
 }
 
+
 void evdi_close(evdi_handle handle)
 {
 	if (handle != EVDI_INVALID_HANDLE) {
 		close(handle->fd);
 		free(handle);
+		for (int device_index = 0; device_index < EVDI_USAGE_LEN; device_index++) {
+			if (card_usage[device_index] == handle) {
+				card_usage[device_index] = EVDI_INVALID_HANDLE;
+				evdi_log("Marking /dev/dri/card%d as unused", device_index);
+			}
+		}
 	}
 }
 
 void evdi_connect(evdi_handle handle,
 		  const unsigned char *edid,
 		  const unsigned int edid_length,
-		  const uint32_t sku_area_limit)
+		  const uint32_t pixel_area_limit,
+		  const uint32_t pixel_per_second_limit)
 {
 	struct drm_evdi_connect cmd = {
 		.connected = 1,
 		.dev_index = handle->device_index,
 		.edid = edid,
 		.edid_length = edid_length,
-		.sku_area_limit = sku_area_limit,
+		.pixel_area_limit = pixel_area_limit,
+		.pixel_per_second_limit = pixel_per_second_limit,
 	};
 
 	do_ioctl(handle->fd, DRM_IOCTL_EVDI_CONNECT, &cmd, "connect");
@@ -533,44 +692,22 @@ void evdi_connect(evdi_handle handle,
 
 void evdi_disconnect(evdi_handle handle)
 {
-	struct drm_evdi_connect cmd = { 0, 0, 0, 0, 0 };
+	struct drm_evdi_connect cmd = { 0, 0, 0, 0, 0, 0 };
 
 	do_ioctl(handle->fd, DRM_IOCTL_EVDI_CONNECT, &cmd, "disconnect");
 }
 
-void evdi_enable_cursor_events(evdi_handle handle)
+void evdi_enable_cursor_events(evdi_handle handle, bool enable)
 {
-	char path[PATH_MAX] = {0};
-	static const char enable[] = "Y";
-	int path_len = 0;
-	FILE *cursor_evs = NULL;
-	size_t written = 0;
-	const size_t elem_bytes = 1;
-	int errcode = 0;
+	struct drm_evdi_enable_cursor_events cmd = {
+		.enable = enable,
+	};
 
-	if (evdi_device_to_platform(handle->device_index, path) !=
-	    AVAILABLE) {
-		evdi_log("Failed to enable cursor events");
-		evdi_log("Device /dev/dri/card%d, device is not available.",
-			handle->device_index);
-		return;
-	}
+	evdi_log("%s cursor events on /dev/dri/card%d",
+		(enable ? "Enabling" : "Disabling"),
+		handle->device_index);
 
-	path_len = strlen(path);
-	snprintf(path+path_len, PATH_MAX-path_len, "/cursor_events");
-	cursor_evs = fopen(path, "w");
-	if (cursor_evs == NULL) {
-		evdi_log("Failed to open %s, err: %s", path, strerror(errno));
-		return;
-	}
-
-	written = fwrite(enable, elem_bytes, sizeof(enable), cursor_evs);
-	errcode = errno;
-	fclose(cursor_evs);
-	evdi_log("Enabling cursor events on /dev/dri/card%d %s %s",
-		handle->device_index,
-		written < sizeof(enable) ? "failed: " : "succeeded",
-		written < sizeof(enable) ? strerror(errcode) : "");
+	do_ioctl(handle->fd, DRM_IOCTL_EVDI_ENABLE_CURSOR_EVENTS, &cmd, "enable cursor events");
 }
 
 void evdi_grab_pixels(evdi_handle handle,
@@ -664,6 +801,20 @@ bool evdi_request_update(evdi_handle handle, int bufferId)
 	}
 }
 
+void evdi_ddcci_response(evdi_handle handle, const unsigned char *buffer,
+			const uint32_t buffer_length,
+			const bool result)
+{
+	struct drm_evdi_ddcci_response cmd = {
+		.buffer = buffer,
+		.buffer_length = buffer_length,
+		.result = result,
+	};
+
+	do_ioctl(handle->fd, DRM_IOCTL_EVDI_DDCCI_RESPONSE, &cmd,
+		"ddcci_response");
+}
+
 static struct evdi_mode to_evdi_mode(struct drm_evdi_event_mode_changed *event)
 {
 	struct evdi_mode mode;
@@ -714,6 +865,8 @@ static struct evdi_cursor_set to_evdi_cursor_set(
 			cursor_set.buffer = malloc(size);
 			memcpy(cursor_set.buffer, ptr, size);
 			munmap(ptr, size);
+		} else {
+			evdi_log("Error: mmap failed with error: %s", strerror(errno));
 		}
 	}
 
@@ -731,6 +884,19 @@ static struct evdi_cursor_move to_evdi_cursor_move(
 	return cursor_move;
 }
 
+static struct evdi_ddcci_data to_evdi_ddcci_data(
+		struct drm_evdi_event_ddcci_data *event)
+{
+	struct evdi_ddcci_data ddcci_data;
+
+	ddcci_data.address = event->address;
+	ddcci_data.flags = event->flags;
+	ddcci_data.buffer = &event->buffer[0];
+	ddcci_data.buffer_length = event->buffer_length;
+
+	return ddcci_data;
+}
+
 static void evdi_handle_event(evdi_handle handle,
 			      struct evdi_event_context *evtctx,
 			      struct drm_event *e)
@@ -776,10 +942,13 @@ static void evdi_handle_event(evdi_handle handle,
 		if (evtctx->cursor_set_handler) {
 			struct drm_evdi_event_cursor_set *event =
 				(struct drm_evdi_event_cursor_set *) e;
+			struct evdi_cursor_set cursor_set = to_evdi_cursor_set(handle, event);
 
-			evtctx->cursor_set_handler(to_evdi_cursor_set(handle,
-								      event),
-						   evtctx->user_data);
+			if (cursor_set.enabled && cursor_set.buffer == NULL)
+				evdi_log("Error: Cursor buffer is null!");
+			else
+				evtctx->cursor_set_handler(cursor_set,
+							   evtctx->user_data);
 		}
 		break;
 
@@ -793,6 +962,16 @@ static void evdi_handle_event(evdi_handle handle,
 		}
 		break;
 
+	case DRM_EVDI_EVENT_DDCCI_DATA:
+		if (evtctx->ddcci_data_handler) {
+			struct drm_evdi_event_ddcci_data *event =
+				(struct drm_evdi_event_ddcci_data *) e;
+
+			evtctx->ddcci_data_handler(to_evdi_ddcci_data(event),
+						   evtctx->user_data);
+		}
+		break;
+
 	default:
 		evdi_log("Warning: Unhandled event");
 	}
