@@ -12,7 +12,7 @@ extern "C" {
 #endif
 
 #define LIBEVDI_VERSION_MAJOR 1
-#define LIBEVDI_VERSION_MINOR 9
+#define LIBEVDI_VERSION_MINOR 12
 #define LIBEVDI_VERSION_PATCH 0
 
 struct evdi_lib_version {
@@ -107,9 +107,10 @@ evdi_handle evdi_open_attached_to(const char *sysfs_parent_device);
 void evdi_close(evdi_handle handle);
 void evdi_connect(evdi_handle handle, const unsigned char *edid,
 		  const unsigned int edid_length,
-		  const uint32_t sku_area_limit);
+		  const uint32_t pixel_area_limit,
+		  const uint32_t pixel_per_second_limit);
 void evdi_disconnect(evdi_handle handle);
-void evdi_enable_cursor_events(evdi_handle handle);
+void evdi_enable_cursor_events(evdi_handle handle, bool enable);
 
 void evdi_grab_pixels(evdi_handle handle,
 		      struct evdi_rect *rects,
