@@ -12,13 +12,13 @@
  */
 
 #include <linux/version.h>
-#if KERNEL_VERSION(5, 5, 0) <= LINUX_VERSION_CODE
+#if KERNEL_VERSION(5, 5, 0) <= LINUX_VERSION_CODE || defined(EL8)
 #else
 #include <drm/drmP.h>
 #endif
 #include <drm/drm_crtc.h>
 #include <drm/drm_crtc_helper.h>
-#include "evdi_drv.h"
+#include "evdi_drm_drv.h"
 
 /* dummy encoder */
 static void evdi_enc_destroy(struct drm_encoder *encoder)
