@@ -12,7 +12,11 @@
  */
 
 #include <linux/version.h>
-#if KERNEL_VERSION(5, 5, 0) <= LINUX_VERSION_CODE
+#if KERNEL_VERSION(5, 16, 0) <= LINUX_VERSION_CODE
+#include <drm/drm_vblank.h>
+#include <drm/drm_damage_helper.h>
+#elif KERNEL_VERSION(5, 0, 0) <= LINUX_VERSION_CODE || defined(EL8)
+#include <drm/drm_damage_helper.h>
 #else
 #include <drm/drmP.h>
 #endif
@@ -22,10 +26,14 @@
 #include <drm/drm_plane_helper.h>
 #include <drm/drm_atomic_helper.h>
 #include "evdi_drm.h"
-#include "evdi_drv.h"
+#include "evdi_drm_drv.h"
 #include "evdi_cursor.h"
 #include "evdi_params.h"
+#if KERNEL_VERSION(5, 13, 0) <= LINUX_VERSION_CODE
+#include <drm/drm_gem_atomic_helper.h>
+#else
 #include <drm/drm_gem_framebuffer_helper.h>
+#endif
 
 static void evdi_crtc_dpms(__always_unused struct drm_crtc *crtc,
 			   __always_unused int mode)
@@ -36,6 +44,7 @@ static void evdi_crtc_dpms(__always_unused struct drm_crtc *crtc,
 static void evdi_crtc_disable(__always_unused struct drm_crtc *crtc)
 {
 	EVDI_CHECKPT();
+	drm_crtc_vblank_off(crtc);
 }
 
 static void evdi_crtc_destroy(struct drm_crtc *crtc)
@@ -56,36 +65,44 @@ static void evdi_crtc_set_nofb(__always_unused struct drm_crtc *crtc)
 
 static void evdi_crtc_atomic_flush(
 	struct drm_crtc *crtc
+#if KERNEL_VERSION(5, 11, 0) <= LINUX_VERSION_CODE || defined(RPI) || defined(EL8)
+	, struct drm_atomic_state *state
+#else
 	, __always_unused struct drm_crtc_state *old_state
+#endif
 	)
 {
-	struct drm_crtc_state *state = crtc->state;
+#if KERNEL_VERSION(5, 11, 0) <= LINUX_VERSION_CODE || defined(RPI) || defined(EL8)
+	struct drm_crtc_state *crtc_state = drm_atomic_get_new_crtc_state(state, crtc);
+#else
+	struct drm_crtc_state *crtc_state = crtc->state;
+#endif
 	struct evdi_device *evdi = crtc->dev->dev_private;
-	unsigned long flags;
+	bool notify_mode_changed = crtc_state->active &&
+				   (crtc_state->mode_changed || evdi_painter_needs_full_modeset(evdi->painter));
+	bool notify_dpms = crtc_state->active_changed || evdi_painter_needs_full_modeset(evdi->painter);
 
-	if (state->event) {
-		spin_lock_irqsave(&crtc->dev->event_lock, flags);
-		drm_crtc_send_vblank_event(crtc, state->event);
-		state->event = NULL;
-		spin_unlock_irqrestore(&crtc->dev->event_lock, flags);
-	}
-	if (state->mode_changed && state->active)
-		evdi_painter_mode_changed_notify(evdi, &state->adjusted_mode);
+	if (notify_mode_changed)
+		evdi_painter_mode_changed_notify(evdi, &crtc_state->adjusted_mode);
 
-	if (state->active_changed)
+	if (notify_dpms)
 		evdi_painter_dpms_notify(evdi,
-			state->active ? DRM_MODE_DPMS_ON : DRM_MODE_DPMS_OFF);
+			crtc_state->active ? DRM_MODE_DPMS_ON : DRM_MODE_DPMS_OFF);
 
-	evdi_painter_send_update_ready_if_needed(evdi);
+	evdi_painter_set_vblank(evdi->painter, crtc, crtc_state->event);
+	evdi_painter_send_update_ready_if_needed(evdi->painter);
+	crtc_state->event = NULL;
 }
 
+#if KERNEL_VERSION(5, 10, 0) <= LINUX_VERSION_CODE || defined(EL8)
+#else
 static void evdi_mark_full_screen_dirty(struct evdi_device *evdi)
 {
 	const struct drm_clip_rect rect =
 		evdi_painter_framebuffer_size(evdi->painter);
 
 	evdi_painter_mark_dirty(evdi, &rect);
-	evdi_painter_send_update_ready_if_needed(evdi);
+	evdi_painter_send_update_ready_if_needed(evdi->painter);
 }
 
 static int evdi_crtc_cursor_set(struct drm_crtc *crtc,
@@ -129,8 +146,11 @@ static int evdi_crtc_cursor_set(struct drm_crtc *crtc,
 	evdi_cursor_set(evdi->cursor,
 			eobj, width, height, hot_x, hot_y,
 			format, stride);
-
+#if KERNEL_VERSION(5, 9, 0) <= LINUX_VERSION_CODE || defined(EL8)
+	drm_gem_object_put(obj);
+#else
 	drm_gem_object_put_unlocked(obj);
+#endif
 
 	/*
 	 * For now we don't care whether the application wanted the mouse set,
@@ -158,6 +178,7 @@ static int evdi_crtc_cursor_move(struct drm_crtc *crtc, int x, int y)
 
 	return 0;
 }
+#endif
 
 static struct drm_crtc_helper_funcs evdi_helper_funcs = {
 	.mode_set_nofb  = evdi_crtc_set_nofb,
@@ -168,6 +189,17 @@ static struct drm_crtc_helper_funcs evdi_helper_funcs = {
 	.disable        = evdi_crtc_disable
 };
 
+#if KERNEL_VERSION(5, 11, 0) <= LINUX_VERSION_CODE || defined(RPI) || defined(EL8)
+static int evdi_enable_vblank(__always_unused struct drm_crtc *crtc)
+{
+	return 1;
+}
+
+static void evdi_disable_vblank(__always_unused struct drm_crtc *crtc)
+{
+}
+#endif
+
 static const struct drm_crtc_funcs evdi_crtc_funcs = {
 	.reset                  = drm_atomic_helper_crtc_reset,
 	.destroy                = evdi_crtc_destroy,
@@ -176,17 +208,40 @@ static const struct drm_crtc_funcs evdi_crtc_funcs = {
 	.atomic_duplicate_state = drm_atomic_helper_crtc_duplicate_state,
 	.atomic_destroy_state   = drm_atomic_helper_crtc_destroy_state,
 
+#if KERNEL_VERSION(5, 10, 0) <= LINUX_VERSION_CODE || defined(EL8)
+#else
 	.cursor_set2            = evdi_crtc_cursor_set,
-	.cursor_move            = evdi_crtc_cursor_move
+	.cursor_move            = evdi_crtc_cursor_move,
+#endif
+#if KERNEL_VERSION(5, 11, 0) <= LINUX_VERSION_CODE || defined(RPI) || defined(EL8)
+	.enable_vblank          = evdi_enable_vblank,
+	.disable_vblank         = evdi_disable_vblank,
+#endif
 };
 
 static void evdi_plane_atomic_update(struct drm_plane *plane,
-				     struct drm_plane_state *old_state)
+#if KERNEL_VERSION(5, 13, 0) <= LINUX_VERSION_CODE
+				     struct drm_atomic_state *atom_state
+#else
+				     struct drm_plane_state *old_state
+#endif
+		)
 {
+#if KERNEL_VERSION(5, 13, 0) <= LINUX_VERSION_CODE
+	struct drm_plane_state *old_state = drm_atomic_get_old_plane_state(atom_state, plane);
+#else
+#endif
 	struct drm_plane_state *state;
 	struct evdi_device *evdi;
+	struct evdi_painter *painter;
 	struct drm_crtc *crtc;
 
+#if KERNEL_VERSION(5, 0, 0) <= LINUX_VERSION_CODE || defined(EL8)
+	struct drm_atomic_helper_damage_iter iter;
+	struct drm_rect rect;
+	struct drm_clip_rect clip_rect;
+#endif
+
 	if (!plane || !plane->state) {
 		EVDI_WARN("Plane state is null\n");
 		return;
@@ -199,6 +254,7 @@ static void evdi_plane_atomic_update(struct drm_plane *plane,
 
 	state = plane->state;
 	evdi = plane->dev->dev_private;
+	painter = evdi->painter;
 	crtc = state->crtc;
 
 	if (!old_state->crtc && state->crtc)
@@ -216,21 +272,39 @@ static void evdi_plane_atomic_update(struct drm_plane *plane,
 		};
 
 		if (!old_fb && crtc)
-			evdi_painter_force_full_modeset(evdi);
+			evdi_painter_force_full_modeset(painter);
 
 		if (old_fb &&
 		    fb->format && old_fb->format &&
 		    fb->format->format != old_fb->format->format)
-			evdi_painter_force_full_modeset(evdi);
+			evdi_painter_force_full_modeset(painter);
 
 		if (fb != old_fb ||
-		    evdi_painter_needs_full_modeset(evdi)) {
+		    evdi_painter_needs_full_modeset(painter)) {
 
-			evdi_painter_set_scanout_buffer(evdi, efb);
+			evdi_painter_set_scanout_buffer(painter, efb);
+
+#if KERNEL_VERSION(5, 0, 0) <= LINUX_VERSION_CODE || defined(EL8)
+			state->visible = true;
+			state->src.x1 = 0;
+			state->src.y1 = 0;
+			state->src.x2 = fb->width << 16;
+			state->src.y2 = fb->height << 16;
+
+			drm_atomic_helper_damage_iter_init(&iter, old_state, state);
+			while (drm_atomic_helper_damage_iter_next(&iter, &rect)) {
+				clip_rect.x1 = rect.x1;
+				clip_rect.y1 = rect.y1;
+				clip_rect.x2 = rect.x2;
+				clip_rect.y2 = rect.y2;
+				evdi_painter_mark_dirty(evdi, &clip_rect);
+			}
+#endif
+
+		};
+
+		if (evdi_painter_get_num_dirts(painter) == 0)
 			evdi_painter_mark_dirty(evdi, &fullscreen_rect);
-		} else if (evdi_painter_get_num_dirts(evdi) == 0) {
-			evdi_painter_mark_dirty(evdi, &fullscreen_rect);
-		}
 	}
 }
 
@@ -244,8 +318,17 @@ static void evdi_cursor_atomic_get_rect(struct drm_clip_rect *rect,
 }
 
 static void evdi_cursor_atomic_update(struct drm_plane *plane,
-				      struct drm_plane_state *old_state)
+#if KERNEL_VERSION(5, 13, 0) <= LINUX_VERSION_CODE
+				     struct drm_atomic_state *atom_state
+#else
+				     struct drm_plane_state *old_state
+#endif
+		)
 {
+#if KERNEL_VERSION(5, 13, 0) <= LINUX_VERSION_CODE
+	struct drm_plane_state *old_state = drm_atomic_get_old_plane_state(atom_state, plane);
+#else
+#endif
 	if (plane && plane->state && plane->dev && plane->dev->dev_private) {
 		struct drm_plane_state *state = plane->state;
 		struct evdi_device *evdi = plane->dev->dev_private;
@@ -287,13 +370,20 @@ static void evdi_cursor_atomic_update(struct drm_plane *plane,
 
 		mutex_unlock(&plane->dev->struct_mutex);
 		if (!evdi->cursor_events_enabled) {
-			evdi_cursor_atomic_get_rect(&old_rect, old_state);
-			evdi_cursor_atomic_get_rect(&rect, state);
+			if (fb != NULL) {
+				if (efb->obj->allow_sw_cursor_rect_updates) {
+					evdi_cursor_atomic_get_rect(&old_rect, old_state);
+					evdi_cursor_atomic_get_rect(&rect, state);
 
-			evdi_painter_mark_dirty(evdi, &old_rect);
-			evdi_painter_mark_dirty(evdi, &rect);
+					evdi_painter_mark_dirty(evdi, &old_rect);
+				} else {
+					rect = evdi_painter_framebuffer_size(evdi->painter);
+				}
+				evdi_painter_mark_dirty(evdi, &rect);
+			}
 			return;
 		}
+
 		if (cursor_changed)
 			evdi_painter_send_cursor_set(evdi->painter,
 						     evdi->cursor);
@@ -305,12 +395,20 @@ static void evdi_cursor_atomic_update(struct drm_plane *plane,
 
 static const struct drm_plane_helper_funcs evdi_plane_helper_funcs = {
 	.atomic_update = evdi_plane_atomic_update,
+#if KERNEL_VERSION(5, 13, 0) <= LINUX_VERSION_CODE
+	.prepare_fb = drm_gem_plane_helper_prepare_fb
+#else
 	.prepare_fb = drm_gem_fb_prepare_fb
+#endif
 };
 
 static const struct drm_plane_helper_funcs evdi_cursor_helper_funcs = {
 	.atomic_update = evdi_cursor_atomic_update,
+#if KERNEL_VERSION(5, 13, 0) <= LINUX_VERSION_CODE
+	.prepare_fb = drm_gem_plane_helper_prepare_fb
+#else
 	.prepare_fb = drm_gem_fb_prepare_fb
+#endif
 };
 
 static const struct drm_plane_funcs evdi_plane_funcs = {
@@ -336,10 +434,11 @@ static struct drm_plane *evdi_create_plane(
 {
 	struct drm_plane *plane;
 	int ret;
+	char *plane_type = (type == DRM_PLANE_TYPE_CURSOR) ? "cursor" : "primary";
 
 	plane = kzalloc(sizeof(*plane), GFP_KERNEL);
 	if (plane == NULL) {
-		EVDI_ERROR("Failed to allocate primary plane\n");
+		EVDI_ERROR("Failed to allocate %s plane\n", plane_type);
 		return NULL;
 	}
 	plane->format_default = true;
@@ -356,7 +455,7 @@ static struct drm_plane *evdi_create_plane(
 				       );
 
 	if (ret) {
-		EVDI_ERROR("Failed to initialize primary plane\n");
+		EVDI_ERROR("Failed to initialize %s plane\n", plane_type);
 		kfree(plane);
 		return NULL;
 	}
@@ -380,6 +479,16 @@ static int evdi_crtc_init(struct drm_device *dev)
 
 	primary_plane = evdi_create_plane(dev, DRM_PLANE_TYPE_PRIMARY,
 					  &evdi_plane_helper_funcs);
+
+#if KERNEL_VERSION(5, 10, 0) <= LINUX_VERSION_CODE || defined(EL8)
+	cursor_plane = evdi_create_plane(dev, DRM_PLANE_TYPE_CURSOR,
+						&evdi_cursor_helper_funcs);
+#endif
+
+#if KERNEL_VERSION(5, 0, 0) <= LINUX_VERSION_CODE || defined(EL8)
+	drm_plane_enable_fb_damage_clips(primary_plane);
+#endif
+
 	status = drm_crtc_init_with_planes(dev, crtc,
 					   primary_plane, cursor_plane,
 					   &evdi_crtc_funcs,
@@ -392,29 +501,11 @@ static int evdi_crtc_init(struct drm_device *dev)
 	return 0;
 }
 
-static int evdi_atomic_check(struct drm_device *dev,
-			     struct drm_atomic_state *state)
-{
-	struct drm_crtc *crtc;
-	struct drm_crtc_state *crtc_state;
-	int i;
-	struct evdi_device *evdi = dev->dev_private;
-
-	if (evdi_painter_needs_full_modeset(evdi)) {
-		for_each_new_crtc_in_state(state, crtc, crtc_state, i) {
-			crtc_state->active_changed = true;
-			crtc_state->mode_changed = true;
-		}
-	}
-
-	return drm_atomic_helper_check(dev, state);
-}
-
 static const struct drm_mode_config_funcs evdi_mode_funcs = {
 	.fb_create = evdi_fb_user_fb_create,
 	.output_poll_changed = NULL,
 	.atomic_commit = drm_atomic_helper_commit,
-	.atomic_check = evdi_atomic_check
+	.atomic_check = drm_atomic_helper_check
 };
 
 void evdi_modeset_init(struct drm_device *dev)
@@ -422,13 +513,14 @@ void evdi_modeset_init(struct drm_device *dev)
 	struct drm_encoder *encoder;
 
 	EVDI_CHECKPT();
+
 	drm_mode_config_init(dev);
 
 	dev->mode_config.min_width = 64;
 	dev->mode_config.min_height = 64;
 
-	dev->mode_config.max_width = 3840;
-	dev->mode_config.max_height = 2160;
+	dev->mode_config.max_width = 7680;
+	dev->mode_config.max_height = 4320;
 
 	dev->mode_config.prefer_shadow = 0;
 	dev->mode_config.preferred_depth = 24;
@@ -444,8 +536,10 @@ void evdi_modeset_init(struct drm_device *dev)
 	drm_mode_config_reset(dev);
 }
 
-void evdi_modeset_cleanup(struct drm_device *dev)
+void evdi_modeset_cleanup(__maybe_unused struct drm_device *dev)
 {
-	EVDI_CHECKPT();
+#if KERNEL_VERSION(5, 8, 0) <= LINUX_VERSION_CODE
+#else
 	drm_mode_config_cleanup(dev);
+#endif
 }
