@@ -10,7 +10,11 @@
 #include "linux/thread_info.h"
 #include "linux/mm.h"
 #include <linux/version.h>
-#if KERNEL_VERSION(5, 5, 0) <= LINUX_VERSION_CODE || defined(EL8)
+#if KERNEL_VERSION(5, 16, 0) <= LINUX_VERSION_CODE
+#include <drm/drm_file.h>
+#include <drm/drm_vblank.h>
+#include <drm/drm_ioctl.h>
+#elif KERNEL_VERSION(5, 5, 0) <= LINUX_VERSION_CODE || defined(EL8)
 #else
 #include <drm/drmP.h>
 #endif
@@ -27,6 +31,10 @@
 
 #include <linux/dma-buf.h>
 
+#if KERNEL_VERSION(5, 16, 0) <= LINUX_VERSION_CODE
+MODULE_IMPORT_NS(DMA_BUF);
+#endif
+
 #if KERNEL_VERSION(5, 1, 0) <= LINUX_VERSION_CODE || defined(EL8)
 #include <drm/drm_probe_helper.h>
 #endif
@@ -239,7 +247,6 @@ u8 *evdi_painter_get_edid_copy(struct evdi_device *evdi)
 			memcpy(block,
 			       evdi->painter->edid,
 			       evdi->painter->edid_length);
-			EVDI_DEBUG("(dev=%d) EDID valid\n", evdi->dev_index);
 		}
 	}
 	painter_unlock(evdi->painter);
@@ -318,7 +325,7 @@ static void evdi_painter_send_event(struct evdi_painter *painter,
 	}
 
 	if (!painter->drm_filp) {
-		EVDI_WARN("Painter is not connected!");
+		EVDI_VERBOSE("Painter is not connected!");
 		drm_event_cancel_free(painter->drm_device, event);
 		return;
 	}
@@ -496,31 +503,6 @@ static void evdi_painter_send_dpms(struct evdi_painter *painter, int mode)
 	evdi_painter_send_event(painter, event);
 }
 
-static struct drm_pending_event *create_crtc_state_event(int state)
-{
-	struct evdi_event_crtc_state_pending *event;
-
-	event = kzalloc(sizeof(*event), GFP_KERNEL);
-	if (!event) {
-		EVDI_ERROR("Failed to create crtc state event");
-		return NULL;
-	}
-
-	event->crtc_state.base.type = DRM_EVDI_EVENT_CRTC_STATE;
-	event->crtc_state.base.length = sizeof(event->crtc_state);
-	event->crtc_state.state = state;
-	event->base.event = &event->crtc_state.base;
-	return &event->base;
-}
-
-static void evdi_painter_send_crtc_state(struct evdi_painter *painter,
-					 int state)
-{
-	struct drm_pending_event *event = create_crtc_state_event(state);
-
-	evdi_painter_send_event(painter, event);
-}
-
 static struct drm_pending_event *create_mode_changed_event(
 	struct drm_display_mode *current_mode,
 	int32_t bits_per_pixel,
@@ -591,7 +573,8 @@ struct drm_clip_rect evdi_painter_framebuffer_size(
 	painter_lock(painter);
 	efb = painter->scanout_fb;
 	if (!efb) {
-		EVDI_DEBUG("Scanout buffer not set.");
+		if (painter->is_connected)
+			EVDI_WARN("Scanout buffer not set.");
 		goto unlock;
 	}
 	rect.x1 = 0;
@@ -618,14 +601,15 @@ void evdi_painter_mark_dirty(struct evdi_device *evdi,
 	painter_lock(painter);
 	efb = painter->scanout_fb;
 	if (!efb) {
-		EVDI_DEBUG("(dev=%d) Skip clip rect. Scanout buffer not set.\n",
+		if (painter->is_connected)
+			EVDI_WARN("(card%d) Skip clip rect. Scanout buffer not set.\n",
 			   evdi->dev_index);
 		goto unlock;
 	}
 
 	rect = evdi_framebuffer_sanitize_rect(efb, dirty_rect);
 
-	EVDI_VERBOSE("(dev=%d) %d,%d-%d,%d\n", evdi->dev_index, rect.x1,
+	EVDI_VERBOSE("(card%d) %d,%d-%d,%d\n", evdi->dev_index, rect.x1,
 		     rect.y1, rect.x2, rect.y2);
 
 	if (painter->num_dirts == MAX_DIRTS)
@@ -695,8 +679,11 @@ void evdi_painter_send_update_ready_if_needed(struct evdi_painter *painter)
 	EVDI_CHECKPT();
 	if (painter) {
 		painter_lock(painter);
-
+#if KERNEL_VERSION(5, 10, 0) <= LINUX_VERSION_CODE || defined(EL8)
+		if (painter->was_update_requested && painter->num_dirts) {
+#else
 		if (painter->was_update_requested) {
+#endif
 			evdi_painter_send_update_ready(painter);
 			painter->was_update_requested = false;
 		}
@@ -707,38 +694,47 @@ void evdi_painter_send_update_ready_if_needed(struct evdi_painter *painter)
 	}
 }
 
+static const char * const dpms_str[] = { "on", "standby", "suspend", "off" };
+
 void evdi_painter_dpms_notify(struct evdi_device *evdi, int mode)
 {
 	struct evdi_painter *painter = evdi->painter;
+	const char *mode_str;
 
-	if (painter) {
-		EVDI_DEBUG("(dev=%d) Notifying dpms mode: %d\n",
-			   evdi->dev_index, mode);
-		evdi_painter_send_dpms(painter, mode);
-	} else {
-		EVDI_WARN("Painter does not exist!");
+	if (!painter) {
+		EVDI_WARN("(card%d) Painter does not exist!", evdi->dev_index);
+		return;
 	}
+
+	if (!painter->is_connected)
+		return;
+
+	switch (mode) {
+	case DRM_MODE_DPMS_ON:
+	case DRM_MODE_DPMS_STANDBY:
+	case DRM_MODE_DPMS_SUSPEND:
+	case DRM_MODE_DPMS_OFF:
+		mode_str = dpms_str[mode];
+		break;
+	default:
+		mode_str = "unknown";
+	};
+	EVDI_INFO("(card%d) Notifying display power state: %s",
+		   evdi->dev_index, mode_str);
+	evdi_painter_send_dpms(painter, mode);
 }
 
-void evdi_painter_crtc_state_notify(struct evdi_device *evdi, int state)
-{
-	struct evdi_painter *painter = evdi->painter;
-
-	if (painter) {
-		EVDI_DEBUG("(dev=%d) Notifying crtc state: %d\n",
-			   evdi->dev_index, state);
-		evdi_painter_send_crtc_state(painter, state);
-	} else {
-		EVDI_WARN("Painter does not exist!");
-	}
-}
-
-static void evdi_log_pixel_format(uint32_t pixel_format)
+static void evdi_log_pixel_format(uint32_t pixel_format,
+		char *buf, size_t size)
 {
+#if KERNEL_VERSION(5, 14, 0) <= LINUX_VERSION_CODE
+	snprintf(buf, size, "pixel format %p4cc", &pixel_format);
+#else
 	struct drm_format_name_buf format_name;
 
 	drm_get_format_name(pixel_format, &format_name);
-	EVDI_DEBUG("pixel format %s\n", format_name.str);
+	snprintf(buf, size, "pixel format %s", format_name.str);
+#endif
 }
 
 void evdi_painter_mode_changed_notify(struct evdi_device *evdi,
@@ -748,6 +744,7 @@ void evdi_painter_mode_changed_notify(struct evdi_device *evdi,
 	struct drm_framebuffer *fb;
 	int bits_per_pixel;
 	uint32_t pixel_format;
+	char buf[100];
 
 	if (painter == NULL)
 		return;
@@ -759,10 +756,11 @@ void evdi_painter_mode_changed_notify(struct evdi_device *evdi,
 	bits_per_pixel = fb->format->cpp[0] * 8;
 	pixel_format = fb->format->format;
 
-	EVDI_DEBUG("(dev=%d) Notifying mode changed: %dx%d@%d; bpp %d; ",
+
+	evdi_log_pixel_format(pixel_format, buf, sizeof(buf));
+	EVDI_INFO("(card%d) Notifying mode changed: %dx%d@%d; bpp %d; %s",
 		   evdi->dev_index, new_mode->hdisplay, new_mode->vdisplay,
-		   drm_mode_vrefresh(new_mode), bits_per_pixel);
-	evdi_log_pixel_format(pixel_format);
+		   drm_mode_vrefresh(new_mode), bits_per_pixel, buf);
 
 	evdi_painter_send_mode_changed(painter,
 				       new_mode,
@@ -795,7 +793,7 @@ static void evdi_add_i2c_adapter(struct evdi_device *evdi)
 	evdi->i2c_adapter = kzalloc(sizeof(*evdi->i2c_adapter), GFP_KERNEL);
 
 	if (!evdi->i2c_adapter) {
-		EVDI_ERROR("(dev=%d) Failed to allocate for i2c adapter",
+		EVDI_ERROR("(card%d) Failed to allocate for i2c adapter",
 			evdi->dev_index);
 		return;
 	}
@@ -805,19 +803,19 @@ static void evdi_add_i2c_adapter(struct evdi_device *evdi)
 	if (result) {
 		kfree(evdi->i2c_adapter);
 		evdi->i2c_adapter = NULL;
-		EVDI_ERROR("(dev=%d) Failed to add i2c adapter, error %d",
+		EVDI_ERROR("(card%d) Failed to add i2c adapter, error %d",
 			evdi->dev_index, result);
 		return;
 	}
 
-	EVDI_DEBUG("(dev=%d) Added i2c adapter bus number %d",
+	EVDI_INFO("(card%d) Added i2c adapter bus number %d",
 		evdi->dev_index, evdi->i2c_adapter->nr);
 
 	result = sysfs_create_link(&evdi->conn->kdev->kobj,
 			&evdi->i2c_adapter->dev.kobj, "ddc");
 
 	if (result) {
-		EVDI_ERROR("(dev=%d) Failed to create sysfs link, error %d",
+		EVDI_ERROR("(card%d) Failed to create sysfs link, error %d",
 			evdi->dev_index, result);
 		return;
 	}
@@ -826,7 +824,7 @@ static void evdi_add_i2c_adapter(struct evdi_device *evdi)
 static void evdi_remove_i2c_adapter(struct evdi_device *evdi)
 {
 	if (evdi->i2c_adapter) {
-		EVDI_DEBUG("(dev=%d) Removing i2c adapter bus number %d",
+		EVDI_INFO("(card%d) Removing i2c adapter bus number %d",
 			evdi->dev_index, evdi->i2c_adapter->nr);
 
 		sysfs_remove_link(&evdi->conn->kdev->kobj, "ddc");
@@ -841,16 +839,16 @@ static void evdi_remove_i2c_adapter(struct evdi_device *evdi)
 static int
 evdi_painter_connect(struct evdi_device *evdi,
 		     void const __user *edid_data, unsigned int edid_length,
-		     uint32_t sku_area_limit,
-		     struct drm_file *file, int dev_index)
+		     uint32_t pixel_area_limit,
+		     uint32_t pixel_per_second_limit,
+		     struct drm_file *file, __always_unused int dev_index)
 {
 	struct evdi_painter *painter = evdi->painter;
 	struct edid *new_edid = NULL;
-	int expected_edid_size = 0;
+	unsigned int expected_edid_size = 0;
+	char buf[100];
 
-	EVDI_DEBUG("(dev=%d) Process is trying to connect\n",
-		   evdi->dev_index);
-	evdi_log_process();
+	evdi_log_process(buf, sizeof(buf));
 
 	if (edid_length < sizeof(struct edid)) {
 		EVDI_ERROR("Edid length too small\n");
@@ -867,7 +865,7 @@ evdi_painter_connect(struct evdi_device *evdi,
 		return -ENOMEM;
 
 	if (copy_from_user(new_edid, edid_data, edid_length)) {
-		EVDI_ERROR("(dev=%d) Failed to read edid\n", dev_index);
+		EVDI_ERROR("(card%d) Failed to read edid\n", evdi->dev_index);
 		kfree(new_edid);
 		return -EFAULT;
 	}
@@ -882,13 +880,13 @@ evdi_painter_connect(struct evdi_device *evdi,
 	}
 
 	if (painter->drm_filp)
-		EVDI_WARN("(dev=%d) Double connect - replacing %p with %p\n",
-			  dev_index, painter->drm_filp, file);
+		EVDI_WARN("(card%d) Double connect - replacing %p with %p\n",
+			  evdi->dev_index, painter->drm_filp, file);
 
 	painter_lock(painter);
 
-	evdi->dev_index = dev_index;
-	evdi->sku_area_limit = sku_area_limit;
+        evdi->pixel_area_limit = pixel_area_limit;
+	evdi->pixel_per_second_limit = pixel_per_second_limit;
 	painter->drm_filp = file;
 	kfree(painter->edid);
 	painter->edid_length = edid_length;
@@ -896,12 +894,12 @@ evdi_painter_connect(struct evdi_device *evdi,
 	painter->is_connected = true;
 	painter->needs_full_modeset = true;
 
-	evdi_add_i2c_adapter(evdi);
+	if (!evdi->i2c_adapter)
+		evdi_add_i2c_adapter(evdi);
 
 	painter_unlock(painter);
 
-	EVDI_DEBUG("(dev=%d) Connected with %p\n", evdi->dev_index,
-		   painter->drm_filp);
+	EVDI_INFO("(card%d) Connected with %s\n", evdi->dev_index, buf);
 
 	drm_helper_hpd_irq_event(evdi->ddev);
 
@@ -912,6 +910,7 @@ static int evdi_painter_disconnect(struct evdi_device *evdi,
 	struct drm_file *file)
 {
 	struct evdi_painter *painter = evdi->painter;
+	char buf[100];
 
 	EVDI_CHECKPT();
 
@@ -929,8 +928,8 @@ static int evdi_painter_disconnect(struct evdi_device *evdi,
 
 	painter->is_connected = false;
 
-	EVDI_DEBUG("(dev=%d) Disconnected from %p\n", evdi->dev_index,
-		   painter->drm_filp);
+	evdi_log_process(buf, sizeof(buf));
+	EVDI_INFO("(card%d) Disconnected from %s\n", evdi->dev_index, buf);
 	evdi_painter_events_cleanup(painter);
 
 	evdi_painter_send_vblank(painter);
@@ -944,7 +943,6 @@ static int evdi_painter_disconnect(struct evdi_device *evdi,
 	evdi_remove_i2c_adapter(evdi);
 
 	painter->drm_filp = NULL;
-	evdi->dev_index = -1;
 
 	painter->was_update_requested = false;
 	evdi->cursor_events_enabled = false;
@@ -962,10 +960,8 @@ void evdi_painter_close(struct evdi_device *evdi, struct drm_file *file)
 {
 	EVDI_CHECKPT();
 
-	if (evdi->painter)
+	if (evdi->painter && file == evdi->painter->drm_filp)
 		evdi_painter_disconnect(evdi, file);
-	else
-		EVDI_WARN("Painter does not exist!");
 }
 
 int evdi_painter_connect_ioctl(struct drm_device *drm_dev, void *data,
@@ -982,19 +978,20 @@ int evdi_painter_connect_ioctl(struct drm_device *drm_dev, void *data,
 			ret = evdi_painter_connect(evdi,
 					     cmd->edid,
 					     cmd->edid_length,
-					     cmd->sku_area_limit,
+					     cmd->pixel_area_limit,
+					     cmd->pixel_per_second_limit,
 					     file,
 					     cmd->dev_index);
 		else
 			ret = evdi_painter_disconnect(evdi, file);
 
 		if (ret) {
-			EVDI_WARN("(dev=%d)(pid=%d) disconnect failed\n",
+			EVDI_WARN("(card%d)(pid=%d) disconnect failed\n",
 				  evdi->dev_index, (int)task_pid_nr(current));
 		}
 		return ret;
 	}
-	EVDI_WARN("Painter does not exist!");
+	EVDI_WARN("(card%d) Painter does not exist!", evdi->dev_index);
 	return -ENODEV;
 }
 
@@ -1030,7 +1027,7 @@ int evdi_painter_grabpix_ioctl(struct drm_device *drm_dev, void *data,
 	painter_lock(painter);
 
 	if (painter->was_update_requested) {
-		EVDI_WARN("(dev=%d) Update ready not sent,",
+		EVDI_WARN("(card%d) Update ready not sent,",
 			  evdi->dev_index);
 		EVDI_WARN(" but pixels are grabbed.\n");
 	}
@@ -1084,9 +1081,9 @@ int evdi_painter_grabpix_ioctl(struct drm_device *drm_dev, void *data,
 		}
 	}
 
-	if (cmd->buf_width != efb->base.width ||
-		cmd->buf_height != efb->base.height) {
-		EVDI_ERROR("Invalid buffer dimension\n");
+	if ((unsigned int)cmd->buf_width != efb->base.width ||
+		(unsigned int)cmd->buf_height != efb->base.height) {
+		EVDI_DEBUG("Invalid buffer dimension\n");
 		err = -EINVAL;
 		goto err_fb;
 	}
@@ -1149,7 +1146,7 @@ int evdi_painter_request_update_ioctl(struct drm_device *drm_dev,
 
 		if (painter->was_update_requested) {
 			EVDI_WARN
-			  ("(dev=%d) Update was already requested - ignoring\n",
+			  ("(card%d) Update was already requested - ignoring\n",
 			   evdi->dev_index);
 		} else {
 			if (painter->num_dirts > 0)
@@ -1210,6 +1207,9 @@ void evdi_painter_cleanup(struct evdi_painter *painter)
 	kfree(painter->edid);
 	painter->edid_length = 0;
 	painter->edid = NULL;
+	if (painter->scanout_fb)
+		drm_framebuffer_put(&painter->scanout_fb->base);
+	painter->scanout_fb = NULL;
 
 	evdi_painter_send_vblank(painter);
 
@@ -1318,7 +1318,7 @@ bool evdi_painter_i2c_data_notify(struct evdi_painter *painter, struct i2c_msg *
 	}
 
 	if (msg->addr != I2C_ADDRESS_DDCCI) {
-		EVDI_DEBUG("Ignored ddc/ci data for address 0x%x\n", msg->addr);
+		EVDI_WARN("Ignored ddc/ci data for address 0x%x\n", msg->addr);
 		return false;
 	}
 
@@ -1363,3 +1363,14 @@ unlock:
 	painter_unlock(painter);
 	return result;
 }
+
+int evdi_painter_enable_cursor_events_ioctl(struct drm_device *drm_dev, void *data,
+					__always_unused struct drm_file *file)
+{
+	struct evdi_device *evdi = drm_dev->dev_private;
+	struct drm_evdi_enable_cursor_events *cmd = data;
+
+	evdi->cursor_events_enabled = cmd->enable;
+
+	return 0;
+}
