@@ -25,14 +25,14 @@ struct platform_device_info;
 
 #define DRIVER_NAME   "evdi"
 #define DRIVER_DESC   "Extensible Virtual Display Interface"
-#define DRIVER_DATE   "20201210"
+#define DRIVER_DATE   "20220713"
 
 #define DRIVER_MAJOR 1
-#define DRIVER_MINOR 9
+#define DRIVER_MINOR 12
 #define DRIVER_PATCH 0
 
 void evdi_platform_remove_all_devices(struct device *device);
-int evdi_platform_device_count(struct device *device);
+unsigned int evdi_platform_device_count(struct device *device);
 int evdi_platform_add_devices(struct device *device, unsigned int val);
 int evdi_platform_device_add(struct device *device, struct device *parent);
 
